/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.service.impl.util;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

import org.springframework.stereotype.Service;

import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.webgui.common.ClassAndMethodContainer;

/**
 * This tool class can be used to find all plugins and repositories within a given jar file - assuming that the given class loader knows these jars.
 * 
 * @author Nils Christian Ehmke
 */
@Service
public final class PluginFinder {

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public PluginFinder() {
		// No code necessary
	}

	/**
	 * This method delivers all classes which are available in the given jar and are compatible with <code>AbstractRepository</code>..
	 * 
	 * @param url
	 *            The url for the jar.
	 * @param classLoader
	 *            The classloader which should be used to load the classes.
	 * @param classAndMethodContainer
	 *            The container for the necessary reflection methods.
	 * @return A list containing all available repository-classes or null, if an exception occurred.
	 * @throws IOException
	 */
	public List<Class<AbstractRepository>> getAllRepositoriesWithinJar(final URL url, final ClassLoader classLoader,
			final ClassAndMethodContainer classAndMethodContainer) throws IOException {
		// Get a list containing all available classes within the given jar
		final List<Class<?>> clazzes = this.getAllClassesWithinJar(url, classLoader);

		List<Class<AbstractRepository>> result = null;

		result = new ArrayList<Class<AbstractRepository>>();
		for (final Class<?> clazz : clazzes) {
			// This is the cast resulting in an unchecked cast warning.
			if (clazz.isAnnotationPresent(classAndMethodContainer.getRepositoryAnnotationClass())
					&& classAndMethodContainer.getAbstractRepositoryClass().isAssignableFrom(clazz)) {
				result.add((Class<AbstractRepository>) clazz);
			}
		}

		return result;
	}

	/**
	 * This method delivers all classes which are available in the given jar and have the Plugin-Annotation from the Kieker framework (And are correctly compatible
	 * with <code>AbstractPlugin</code>).
	 * 
	 * @param url
	 *            The url for the jar.
	 * @param classLoader
	 *            The classloader which should be used to load the classes.
	 * @param classAndMethodContainer
	 *            The container for the necessary reflection methods.
	 * @return A list containing all available plugin-classes or null, if an exception occurred.
	 * @throws IOException
	 */
	public List<Class<AbstractPlugin>> getAllPluginsWithinJar(final URL url, final ClassLoader classLoader,
			final ClassAndMethodContainer classAndMethodContainer) throws IOException {
		final List<Class<?>> clazzes = this.getAllClassesWithinJar(url, classLoader);
		List<Class<AbstractPlugin>> result = null;

		result = new ArrayList<Class<AbstractPlugin>>();
		for (final Class<?> clazz : clazzes) {
			// This is the cast resulting in an unchecked cast warning.
			if (clazz.isAnnotationPresent(classAndMethodContainer.getPluginAnnotationClass())
					&& classAndMethodContainer.getAbstractPluginClass().isAssignableFrom(clazz)) {
				result.add((Class<AbstractPlugin>) clazz);
			}
		}

		return result;
	}

	/**
	 * This method delivers all classes which are available in the given jar.
	 * 
	 * @param url
	 *            The url for the jar.
	 * @param classLoader
	 *            The classloader which should be used to load the classes.
	 * @return A list containing all available classes or null, if an exception occurred.
	 */
	private List<Class<?>> getAllClassesWithinJar(final URL url, final ClassLoader classLoader) throws IOException {
		JarInputStream stream = null;
		try {
			final List<Class<?>> result = new ArrayList<Class<?>>();
			// Open the jar as a stream and run through all entries within this file
			stream = new JarInputStream(url.openStream());

			JarEntry jarEntry;
			while ((jarEntry = stream.getNextJarEntry()) != null) {
				try {
					// Assemble the correct name
					String name = jarEntry.toString();
					name = name.replace('/', '.');
					name = name.replace(".class", "");
					// Try to find a class with the same name and put it into our list
					final Class<?> c = classLoader.loadClass(name);
					result.add(c);
				} catch (final Throwable ex) { // NOPMD (Generic throwable and empty catch block) NOCS (IllegalCatchCheck)
					// Ignore error.
				}
			}
			// Don't forget to close the jar file again.
			stream.close();
			return result;
		} finally {
			if (stream != null) {
				try {
					stream.close();
				} catch (final IOException ex1) {
					// Ignore this
					ex1.printStackTrace();
				}
			}
		}
	}

}
